/**
	* Ce programme sert  montrer l'utilisation de l'hritage
	* et les classes et mthodes abstraites en Java

	* @author Philippe Langlais
	* @author Mohamed Lokbani (quelques ajustements) 
	* @version 1.1
*/


/**

 *la classe mere de la hierarchie ne contient que deux niveaux
 *on definit ici le comportement par defaut d'un compte.
 *abstract indique que l'on ne peut instancier d'objet de cette classe
 
*/


abstract class Compte {
    protected String nomClient;
    protected String numeroCompte;
    
    /**
       solde en cents
    */
    private long solde;

    /**
       ouverture d'un compte
       @param client pour lequel ce compte sera ouvert
       @param compte numro de compte
       @param montant le montant  l'ouverture
    */

    public Compte(String client, String compte, long montant) {
	nomClient = client;
	numeroCompte = compte;
	solde = montant;
    }

    public String getNom() {
	return nomClient;
    }
    /**
       une mthode abstraite
       pour la mise  jour du compte
    */

    abstract public void maj();

    /**
       depot d'une somme d'argent
       @param montant montant  dposer
    */
    protected void depot(long montant) {
	solde += montant;
    }

    /**
       retrait d'une somme d'argent
       @param montant montant  retirer
    */
    protected void retrait(long montant) {
	solde -= montant;
    }

    /**
       le solde dans le compte
    */
    protected long getSolde() {
      return solde;
    }

    /**
       retourne une chaine que les sous-classes vont utiliser
    */
    public String toString() {
	return nomClient + " solde: " + (solde/100) + "$CA," + Math.abs(solde %100);
    }
}

/**

 *un compte simple n'autorise pas le debit
 *on redfini l'operation de retrait

*/


final class CompteSimple extends Compte {

    /**
       ouverture d'un compte simple sans dbit
       @param client pour lequel ce compte sera ouvert
       @param compte numro de compte
       @param montant le montant  l'ouverture
    */


    public CompteSimple(String client, String compte,long montant) {
	super(client,compte,montant);
    }


    /**
       retrait d'une somme d'argent
       @param montant montant  retirer
    */

    public void retrait (long montant) {
	if (montant < getSolde()) {
	    super.retrait(montant);
	} else
	    System.err.println("Le solde disponible sur le compte " +
			       numeroCompte + " n'est pas suffisant pour le retrait demande");
    }

    /**
       dfinition de la mthode abstraite
       pour la mise  jour du compte
       rien de special a faire ici!
    */

    public void maj() {
	// rien de special a faire ici
    }

    /**
       quelques informations relatives au compte sans crdit
    */    
    public String toString() {
	return "compte sans credit de " + super.toString();
    }
}

/**

 *un compte rmunr
 *on redfinit mthode de mise  jour

*/


final class CompteRemunere extends Compte {

    private double tauxInteret;


    /**
       ouverture d'un compte rmunr
       @param client pour lequel ce compte sera ouvert
       @param compte numro de compte
       @param montant le montant  l'ouverture
       @param tauxInteret taux d'intrt utilis pour la rmunration
    */


    public CompteRemunere(String client, String compte, long montant,double tauxInteret) {
	super(client,compte,montant);
	this.tauxInteret = tauxInteret;
    }

   /**
       dfinition de la mthode abstraite
       pour la mise  jour du compte
       ici on ajoute les intrts
    */
    public void maj() {
	ajouteInteret();
    }
    
    /**
       calcul des intrts
    */
    private boolean ajouteInteret() {
	long s;
	if ((s=getSolde()) > 0) {
	    depot((long) (s * tauxInteret));
	    return true;
	}
	return false; // pas d'interet ajoute
    }
   
    /**
       quelques informations relatives au compte rmunr
    */    
    public String toString() {
	return "compte remunere de " + super.toString();
    }
}


/**

 *un compte crdit
 *le comportement de ce compte est ici celui de la classe
  mre, on ne le modifie pas.
*/


class CompteCredit extends Compte {

    /**
       ouverture d'un compte crdit
       @param client pour lequel ce compte sera ouvert
       @param compte numro de compte
       @param montant le montant  l'ouverture
    */
    public CompteCredit(String client, String compte, long montant) {
	super(client,compte,montant);
    }
   /**
       dfinition de la mthode abstraite
       pour la mise  jour du compte
       rien  faire ici
    */
    public void maj() {
	// rien a faire
    }

    /**
       quelques informations relatives au compte crdit
    */
    public String toString() {
	return "compte avec credit de " + super.toString();
    }
}
/**

 *une banque ayant
 *un nom et un certain nombre de comptes
*/


class Banque {

    private  final int MAX_COMPTES = 1000;
    private  Compte comptes [];
    private int nbComptes;
    private  String nomAgence;

    /**
       le client existe t-il dj dans les comptes ouverts
       @param client le nom du client
    */
    private int existe(String client) {
	for (int i=0; i<nbComptes; i++)
	    if (client.equals(comptes[i].getNom())) return i;
	return -1;
    }

    /**
       retourne le total des soldes des comptes ouverts
    */
    private long totalComptes() {
	long s=0;
	for (int i=0; i<nbComptes; i++)
	    s += comptes[i].getSolde();
	return s;
    }
    /**
       allocation des comptes
       @param agence le nom de l'agence ayant ouvert le compte
    */

    public Banque (String agence) {
	nbComptes = 0;
	comptes = new Compte [MAX_COMPTES];
	nomAgence=agence;
    }

    /**
       ouverture d'un compte rmunr
       @param client pour lequel ce compte sera ouvert
       @param montant le montant  l'ouverture
       @param tauxInteret taux d'intrt utilis pour la rmunration

       le numero de compte est retourne, ainsi sera le code d'acces au compte

    */

    public String ouvertureCompteRemunere(String client, long m,double taux) {
	if ((nbComptes < MAX_COMPTES) && (existe(client) == -1)) {
	    comptes[nbComptes] = new CompteRemunere(client,nomAgence+nbComptes,m,taux);
	    return nomAgence+(nbComptes++);
	}

	System.err.println("Ouverture impossible");
	return "erreur";
    }

   /**
       ouverture d'un compte simple
       @param client pour lequel ce compte sera ouvert
       @param montant le montant  l'ouverture

       le numero de compte est retourne, ainsi sera le code d'acces au compte

    */

    public String ouvertureCompteSimple(String client,long m) {
	if ((nbComptes < MAX_COMPTES) && (existe(client) == -1)) {
	    comptes[nbComptes] = new CompteSimple(client,nomAgence+nbComptes,m);
	    return nomAgence+(nbComptes++);
	}
	System.err.println("Ouverture impossible");
	return "erreur";
    }

   /**
       ouverture d'un compte crdit
       @param client pour lequel ce compte sera ouvert
       @param montant le montant  l'ouverture

       le numero de compte est retourne, ca sera le code d'acces au compte

    */

    public String ouvertureCompteCredit(String client, long m) {
	if ((nbComptes < MAX_COMPTES) && (existe(client) == -1)) {
	    comptes[nbComptes] = new CompteCredit(client,nomAgence+nbComptes,m);
	    return nomAgence+(nbComptes++);
	}
	System.err.println("Ouverture impossible");
	return "erreur";
    }
    
    /**
       methode qui met  jour les comptes. On suppose que cette methode sera appelee
       automatiquement une fois par mois et ce, afin de calculer les interets , etc.
    */

    public void miseAJourMensuelle() {
	for (int i=0; i<nbComptes; i++) comptes[i].maj();
    }

    /**
       un client demande un retrait
       @param client le compte utilis pour le retrait
       @param montant le montant demand
    */
    public void retrait (String client, long montant) {
	int c;
	System.out.println(client + " retrait: " + montant);
	if ((c=existe(client)) != -1)
	    comptes[c].retrait(montant);
	else
	    System.out.println("Client non connu - Transaction interrompue");
    }

    /**
       un client demande un dpt
       @param client le compte utilis pour le dpt
       @param montant la somme dpose
    */
    public void depot (String client, long montant) {
	int c;
	System.out.println(client + " depot: " + montant);
	if ((c=existe(client)) != -1)
	    comptes[c].depot(montant);
	else
	    System.out.println("Client non connu - Transaction interrompue");
    }

    /**
       La liste des comptes crs
    */
    public void listeCompte() {
	for (int i=0; i<nbComptes; i++)
	    System.out.println(comptes[i].toString());
    }
    /**
       affichage de quelques informations relatives  la banque
    */

    public String toString() {
	return "Banque " + nomAgence + " nb comptes: " + nbComptes + " $CA:" + (totalComptes()/100);
    }
}


/**

   classe principale

*/


public class herabst {
    
    static String clientCredit[] = {
	"Charles Dutoit",
	"Robert Lepage",
	"Marie Brassard",
	"Julie Laroche"
    };
    
    static String clientSimple[] = {
	"Myreille Bedard",
	"Cunegonde Ubercule",
    };
    
    static String  clientRemunere[] = {
	"Linda Tremblay",
	"Richard Robinaud",
	"Johanne Thibault"
    };
    
    //Main method
    public static void main(String[] args) {
	long m;
	String Clients [] = new String [100];
	int top=0;
	Banque banque = new Banque("Caisse Pop Desjardin - Cote-des-Neiges");
	
	// ouverture des comptes
	for (int i=0; i<clientCredit.length; i++)
	    System.out.println("compte " +
			       banque.ouvertureCompteCredit(Clients[top++]=clientCredit[i],
							    (long) (Math.random()*10000)) +" ouvert");
	
	for (int i=0; i<clientSimple.length; i++)
	    System.out.println("compte " +
			       banque.ouvertureCompteSimple(Clients[top++]=clientSimple[i],
							    (long) (Math.random()*1000)) + " ouvert");
	
	for (int i=0; i<clientRemunere.length; i++)
	    System.out.println("compte " +
			       banque.ouvertureCompteRemunere(Clients[top++]=clientRemunere[i],
							      (long) (Math.random()*10000), 0.075) + " ouvert");
	
	// quelques operations de dept et de retrait
	int nbOp;
	
	for (int i=0; i<top; i++) {
	    nbOp = 1 + (int) (Math.random() * 4); // nb d'operations pour un client
	    for (int j=0; j<nbOp; j++) {
		m = (long) (Math.random() * 1000); // montant du depot ou retrait
		if ((m % 2) == 0) banque.depot(Clients[i],m);
		else banque.retrait(Clients[i],m);
	    }
	}
	
	System.out.println(banque);
	banque.listeCompte();
	
	// C'est la fin du mois, la banque calcule les intrts des comptes remunrs
	banque.miseAJourMensuelle();
	System.out.println(banque);
	banque.listeCompte();
    }
    
}

